
#include "ContourExtractor.h"
#include "utils\Utils.h"
#include <iostream>

ContourExtractor::ContourExtractor(int width, int height)
	:m_width(width), m_height(height)
{
	m_wf_threshold = 20;

	m_phi_image.Reset(width,height);
	m_grad_phi_image.Reset(width, height);
	m_tmp_phi_image.Reset(width, height);
	m_u_phi_image.Reset(width, height);
	m_v_phi_image.Reset(width, height);

	m_gradient_sigma = 0.3f;
}

void ContourExtractor::WallFollow(const ImgFloat& strength_image, ObjectModel& object_model, Point pt)
{
	WallFollow(strength_image, object_model, &object_model.m_contour_points, pt);
}

void ContourExtractor::WallFollow(const ImgFloat& strength_image, ObjectModel& object_model, vector<Point>* contour_points, Point pt)
{
	try {
	int curr_x, curr_y, start_x, start_y, prev_x, prev_y;
	enum direction_t {UP, LEFT, RIGHT, DOWN};
	direction_t direction = UP;
	int width = strength_image.Width();
	int height = strength_image.Height();
	int prev_size = contour_points->size();
	contour_points->clear();

	int mean_x, mean_y;

	if(pt.x == -1 || pt.y == -1) {
		mean_x = start_x = object_model.m_object_att.m_region_mean[0];
		mean_y = start_y = object_model.m_object_att.m_region_mean[1];
	}
	else {
		mean_x = pt.x;
		mean_y = pt.y;
	}

	// if previous mean was incorrect, then find a new starting position for wall follow 
	// in the neighborhood of the previous mean.
	if(strength_image(mean_x, mean_y) <= 0) {
		//printf("Frame No: %d : Incorrect mean: [%d %d] ", m_frame_no, mean_x, mean_y);
		_GetSeedPoint(mean_x, mean_y, object_model.m_prev_contour_points, strength_image, mean_x, mean_y);
	}

	int iter = 0;
	int cnt = 0;
	float ratio;
	bool repeat = false;

	do {
		contour_points->clear();

		if(cnt % 4 == 0) {
			start_y = mean_y - cnt;
			start_x = mean_x;
			bool flag = false;
			while(strength_image(start_x, start_y) > 0) { start_x--; flag = true; }
			if(flag) 
				start_x++;
			direction = UP;
		}
		else if(cnt % 4 == 1) {
			start_y = mean_y - cnt;
			start_x = mean_x;
			bool flag = false;
			while(strength_image(start_x, start_y) > 0) { start_x++; flag = true; }
			if(flag)
				start_x--;
			direction = DOWN;
		}
		else if(cnt % 4 == 2) {
			start_y = mean_y;
			start_x = mean_x - cnt;
			bool flag = false;
			while(strength_image(start_x, start_y) > 0) { start_y--; flag = true; }
			if(flag)
				start_y++;
			direction = RIGHT;

		}
		else if(cnt % 4 == 3) {
			start_y = mean_y;
			start_x = mean_x - cnt;
			bool flag = false;
			while(strength_image(start_x, start_y) > 0) { start_y++; flag = true; }
			if(flag)
				start_y--;
			direction = LEFT;
		}

		curr_x = start_x; curr_y = start_y;
		prev_x = curr_x; prev_y = curr_y;
		(*contour_points).push_back(Point(curr_x, curr_y));
		iter = 0;
		direction_t start_dir = direction;

		do {
			iter++;
			if(curr_x != prev_x || curr_y != prev_y) {
				(*contour_points).push_back(Point(curr_x, curr_y));
				prev_x = curr_x; prev_y = curr_y;
			}
			if(direction == UP) {
				if(curr_x-1 > 0 && strength_image(curr_x-1, curr_y) > 0) {
					direction = LEFT;
					curr_x--;
				}
				else if(curr_y-1 <= 0 || (curr_y-1 > 0 && strength_image(curr_x, curr_y-1) <= 0))
					direction = RIGHT;
				else
					curr_y--;

			}
			else if(direction == RIGHT) {
				if(curr_y-1 > 0 && strength_image(curr_x,curr_y-1) > 0) {
					direction = UP;
					curr_y--;
				}
				else if(curr_x+1 >= width || (curr_x+1 < width && strength_image(curr_x+1, curr_y) <= 0))
					direction = DOWN;
				else
					curr_x++;
			}
			else if(direction == DOWN) {
				if(curr_x+1 < width && strength_image(curr_x+1, curr_y) > 0) {
					direction = RIGHT;
					curr_x++;
				}
				else if(curr_y+1 >= height || (curr_y+1 < height && strength_image(curr_x, curr_y+1) <= 0))
					direction = LEFT;
				else
					curr_y++;
			}
			else if(direction == LEFT) {
				if(curr_y+1 < height && strength_image(curr_x, curr_y+1) > 0) {
					direction = DOWN;
					curr_y++;
				}
				else if(curr_x-1 <= 0 || (curr_x-1 > 0 && strength_image(curr_x-1, curr_y) <= 0))
					direction = UP;
				else
					curr_x--;
			}

			if(iter > width * height) {
				//cout << "oops !! In Frame no " << m_frame_no << " wall follow went into a loop !" << endl;
				repeat = true;
				break;
			}
		} while(! ( curr_x == start_x && curr_y == start_y && start_dir == direction ) );
		if(repeat) {
			repeat = false;
			cnt++;
			continue;
		}
		cnt++;
		ratio = (float)contour_points->size()/(float)prev_size+0.00000001f;
	} while(cnt < 100 && ratio < 0.5);
	}
	catch(...) {
		cout << "Crashed in WallFollow()" << endl;
	}
}

void ContourExtractor::_GetSeedPoint(int mean_x, int mean_y, const vector<Point>& contour_points, const ImgFloat& strength_image, int& seed_x, int& seed_y)
{
	int i, j;
	int temp_x = -1, temp_y = -1;
	int width = strength_image.Width();
	int height = strength_image.Height();
	int x, y;
	for(i=0; i<contour_points.size(); i++) {
		x = contour_points[i].x; y = contour_points[i].y;
		if(strength_image(x, y) > 0) {
			//printf("Found a new start position for wall follow from prev contour: [%d %d]\n", x, y);
			temp_x = x;
			temp_y = y;
			break;
		}
	}
	if(temp_x != -1 && temp_y != -1) {
		seed_x = temp_x;
		seed_y = temp_y;
		return;
	}

	bool found_object = false;
	int left = (mean_x - m_wf_threshold <= 0) ? 0 : mean_x - m_wf_threshold;
	int right = (mean_x + m_wf_threshold >= width) ? width : mean_x + m_wf_threshold;
	int top = (mean_y - m_wf_threshold <= 0) ? 0 : mean_y - m_wf_threshold;
	int bottom = (mean_y + m_wf_threshold >= height) ? height: mean_y + m_wf_threshold;
	for(i=top; i<bottom; i++) {
		for(j=left; j<right; j++) {
			if(strength_image(j,i) > 0) {
				//printf("Found a new start position for wall follow: [%d %d]\n", j, i);
				seed_x = j; seed_y = i;
				found_object = true;
				break;
			}
		}
		if(found_object) break;
	}
}

void ContourExtractor::DoLevelSets(const ImgBgr& current_frame, const ImgFloat& strength_image, const ImgBinary& object_mask, ObjectModel& object_model, vector<Point> *contour_points)
{
	int time, x, y;

	ImgBgr display_image = current_frame;
	float div_factor = 0.0000000001f;
	float divergence;

	// Initialize phi image
	ImgBinary::ConstIterator p_object_mask;
	ImgFloat::Iterator p_phi_image;
	for(p_object_mask=object_mask.Begin(), p_phi_image=m_phi_image.Begin(); p_object_mask != object_mask.End(); p_object_mask++, p_phi_image++) {
		*p_phi_image = (*p_object_mask) ? 5 : -5;
	}

	// Start Iteration
	for(time=0; time<100; time++) {
		GradMag(m_phi_image, m_gradient_sigma, &m_grad_phi_image);
		Gradient(m_phi_image, m_gradient_sigma, &m_u_phi_image, &m_v_phi_image);

		// Update the phi image
		ImgFloat::ConstIterator p_strength_image = strength_image.Begin();
		ImgFloat::Iterator p_grad_phi_image = m_grad_phi_image.Begin();
		ImgFloat::Iterator p_tmp_phi_image = m_tmp_phi_image.Begin();
		p_phi_image = m_phi_image.Begin();
		int cnt = 0;

		for(y=0; y<m_height; y++) {
			for(x=0; x<m_width; x++, p_strength_image++, p_grad_phi_image++, p_phi_image++, p_tmp_phi_image++) {
				*p_tmp_phi_image = *p_phi_image;
				//if(m_grad_phi_image(x+1,y) != 0 && m_grad_phi_image(x,y) != 0 && m_grad_phi_image(x,y+1) != 0)
				//	divergence = (m_u_phi_image(x+1,y)/abs(m_grad_phi_image(x+1,y))) - (m_u_phi_image(x,y)/abs(m_grad_phi_image(x,y))) + (m_v_phi_image(x,y+1)/abs(m_grad_phi_image(x,y+1))) - (m_v_phi_image(x,y)/abs(m_grad_phi_image(x,y)));
				//else 
					divergence = m_u_phi_image(x+1,y) - m_u_phi_image(x,y) + m_v_phi_image(x,y+1) - m_v_phi_image(x,y);
				*p_phi_image = *p_phi_image + (abs(*p_grad_phi_image) * (*p_strength_image + (div_factor * divergence)));
				//*p_phi_image = *p_phi_image + (*p_strength_image * abs(*p_grad_phi_image));
				if((*p_tmp_phi_image<0 && *p_phi_image>0) || (*p_tmp_phi_image>0 && *p_phi_image<0))
					cnt++;
			}
		}
		
		// Extract the zero level set and display the intermediate results
		
		int i;
		vector<Point> contour_pts;
		cout << "No of sign flips over time: " << cnt << endl;
		_GetSignChanges(m_phi_image, &contour_pts);
		//WallFollow(m_phi_image, object_model, &contour_pts);
		for(i=0; i<contour_pts.size(); i++)
			DrawDot(contour_pts[i], &display_image, Bgr(255,0,0), 2);
		static Figure level_set_wnd("**2");
		level_set_wnd.Draw(display_image);
		display_image = current_frame;
		cout << "No of contour points: " << contour_pts.size() << endl;
		EventLoop();
		

		if(cnt < 10)
			break;
	}

	// Update the contour points based on the final zeroth level set
	WallFollow(m_phi_image, object_model, contour_points);	
}


void ContourExtractor::_GetSignChanges(const ImgFloat& phi_image, vector<Point> *contour_points)
{
	int x, y;
	for(y=1; y<m_height-1; y++) {
		for(x=1; x<m_width-1; x++) {
			if(!((phi_image(x,y)>=0 && phi_image(x-1,y)>=0 && phi_image(x+1,y)>=0 && phi_image(x,y-1)>=0 && phi_image(x,y+1)>=0) || (phi_image(x,y)<=0 && phi_image(x-1,y)<=0 && phi_image(x+1,y)<=0 && phi_image(x,y-1)<=0 && phi_image(x,y+1)<=0)))
				(*contour_points).push_back(Point(x,y));
		}
	}
}

void ContourExtractor::FastLevelSets(const ImgBgr& current_frame, const ImgFloat& strength_image, ImgBinary& mask, vector<Point> *contour_pts)
{
	int i;
	list<Point> frontier;
	int num_pts = contour_pts->size();

	// copy the contour points into frontier
	for(i=0; i<num_pts; i++)
		frontier.push_back((*contour_pts)[i]);

	// Update the frontier with the intermediate points. To accomplish this, find the gradmag of the mask and pick all the positive values
	ImgFloat f_mask(m_width, m_height), f_gradmask(m_width, m_height);
	Convert(mask, &f_mask);
	GradMag(f_mask, 1.5f, &f_gradmask);
	for(int y=0; y<m_height; y++) {
		for(int x=0; x<m_width; x++) {
			if(f_gradmask(x,y) > 0) frontier.push_back(Point(x,y));
		}
	}

	bool flag = false, dilation_check = false, contraction_check = false;
	list<Point>::iterator p_frontier;
	do {
		flag = false;
		for(p_frontier = frontier.begin(); p_frontier != frontier.end(); p_frontier++) {
			Point pt = *p_frontier;
			/* Dilation:
				 Push the neighbor(x) into the queue.
				 mask(neighbor(x)) = 1
				 If no neighbor is added and the current point is surrounded by strength, remove it from queue.
			*/
			dilation_check = false;
			if( pt.x-1 > 0 && !mask(pt.x-1, pt.y) && strength_image(pt.x-1, pt.y) > 0 ) {
				frontier.push_front(Point(pt.x-1,pt.y));			// push_front is very important
				mask(pt.x-1,pt.y) = true;
				flag = dilation_check = true;
			}
			if( pt.x+1 < m_width && !mask(pt.x+1, pt.y) && strength_image(pt.x+1, pt.y) > 0 ) {
				frontier.push_front(Point(pt.x+1,pt.y));			// push_front is very important
				mask(pt.x+1,pt.y) = true;
				flag = dilation_check = true;
			}
			if( pt.y-1 > 0 && !mask(pt.x, pt.y-1) && strength_image(pt.x, pt.y-1) > 0 ) {
				frontier.push_front(Point(pt.x,pt.y-1));			// push_front is very important
				mask(pt.x,pt.y-1) = true;
					flag = dilation_check = true;
			}
			if( pt.y+1 < m_height && !mask(pt.x, pt.y+1) && strength_image(pt.x, pt.y+1) > 0 ) {
				frontier.push_front(Point(pt.x,pt.y+1));			// push_front is very important
				mask(pt.x,pt.y+1) = true;
				flag = dilation_check = true;
			}
			//if(!dilation_check && pt.x-1 > 0 && pt.y-1 > 0 && pt.x+1 < m_width && pt.y+1 < m_height && strength_image(pt.x-1, pt.y) > 0 && strength_image(pt.x+1, pt.y) > 0 && strength_image(pt.x, pt.y-1) > 0 && strength_image(pt.x, pt.y+1) > 0 ) {
			if(pt.x-1 > 0 && pt.y-1 > 0 && pt.x+1 < m_width && pt.y+1 < m_height && strength_image(pt.x-1, pt.y) > 0 && strength_image(pt.x+1, pt.y) > 0 && strength_image(pt.x, pt.y-1) > 0 && strength_image(pt.x, pt.y+1) > 0 ) {
				p_frontier = frontier.erase(p_frontier);
				if(p_frontier != frontier.begin())
					p_frontier--;
			}

			/* Contraction:
				 Push the neighbor(x) into the queue.
				 mask(neighbor(x)) = 0
				 mask(x) = 0
				 If no neighbor is added and if the current point is surrounded by no strength, remove it from queue
			*/
			contraction_check = false;
			if( pt.x-1 > 0 && mask(pt.x-1, pt.y) && strength_image(pt.x-1, pt.y) < 0 ) {
				frontier.push_front(Point(pt.x-1,pt.y));			// push_front is very important
				mask(pt.x-1, pt.y) = false;
				flag = contraction_check = true;
			}
			if( pt.x+1 < m_width && mask(pt.x+1, pt.y) && strength_image(pt.x+1, pt.y) < 0 ) {
				frontier.push_front(Point(pt.x+1,pt.y));			// push_front is very important
				mask(pt.x+1, pt.y) = false;
				flag = contraction_check = true;
			}
			if( pt.y-1 > 0 && mask(pt.x, pt.y-1) && strength_image(pt.x, pt.y-1) < 0 ) {
				frontier.push_front(Point(pt.x,pt.y-1));			// push_front is very important
				mask(pt.x, pt.y-1) = false;
				flag = contraction_check = true;
			}
			if( pt.y+1 < m_height && mask(pt.x, pt.y+1) && strength_image(pt.x, pt.y+1) < 0 ) {
				frontier.push_front(Point(pt.x,pt.y+1));			// push_front is very important
				mask(pt.x, pt.y+1) = false;
				flag = contraction_check = true;
			}
			//if(contraction_check)
			//	mask(pt.x, pt.y) = false;
			//if(!contraction_check && pt.x-1 > 0 && pt.y-1 > 0 && pt.x+1 < m_width && pt.y+1 < m_height && strength_image(pt.x-1, pt.y) < 0 && strength_image(pt.x+1, pt.y) < 0 && strength_image(pt.x, pt.y-1) < 0 && strength_image(pt.x, pt.y+1) < 0 ) {
			if(pt.x-1 > 0 && pt.y-1 > 0 && pt.x+1 < m_width && pt.y+1 < m_height && strength_image(pt.x-1, pt.y) < 0 && strength_image(pt.x+1, pt.y) < 0 && strength_image(pt.x, pt.y-1) < 0 && strength_image(pt.x, pt.y+1) < 0 ) {
				p_frontier = frontier.erase(p_frontier);
				if(p_frontier != frontier.begin())
					p_frontier--;
			}
		}


	} while(flag);

	// copy the frontier elts into contour points
	contour_pts->clear();
	for(p_frontier = frontier.begin(); p_frontier != frontier.end(); p_frontier++)
		(*contour_pts).push_back(*p_frontier);

}