
#ifndef _CONTOUREXTRACTOR_H
#define _CONTOUREXTRACTOR_H

#include "TheTracker.h"
#include "ObjectModel.h"
#include <deque>

class ContourExtractor {
private:
	int m_width, m_height;

	// Wall follow parameters
	int m_wf_threshold;

	double m_gradient_sigma;
	// Scratch images used by level sets.
	ImgFloat m_phi_image, m_grad_phi_image, m_tmp_phi_image;
	ImgFloat m_u_phi_image, m_v_phi_image;

	void _GetSeedPoint(int mean_x, int mean_y, const vector<Point>& contour_points, const ImgFloat& strength_image, int& seed_x, int& seed_y);
	inline void _GetSignChanges(const ImgFloat& phi_image, vector<Point> *contour_points);
public:
	ContourExtractor(int width, int height);

	// Extracts the contour in contour_points passed as an argument
	void WallFollow(const ImgFloat& strength_image, ObjectModel& object_model, vector<Point>* contour_points, Point pt = Point(-1,-1));

	// Extracts the contour and updates the object model contour points.
	void WallFollow(const ImgFloat& strength_image, ObjectModel& object_model, Point pt = Point(-1,-1));

	// Apply the level sets on the current frame starting from previous contour's position
	void DoLevelSets(const ImgBgr& current_frame, const ImgFloat& strength_image, const ImgBinary& object_mask, ObjectModel& object_model, vector<Point> *contour_points);

	// Discrete implementation of level sets using region growing
	void FastLevelSets(const ImgBgr& current_frame, const ImgFloat& strength_image, ImgBinary& mask, vector<Point> *contour_pts);
};

#endif