#ifndef _HOMOGRAPHY2D_H_
#define _HOMOGRAPHY2D_H_

#include "ransacfithomography.h"

int normalize2dpts(double *pts, const int rows, const int npts, double *newpts, double T[9]);
void denormalize(double H[9], const double T1[9], const double T2[9]);
void denormalize_transposed(const double h[9], const double T1[9], const double T2[9], double H[9]);
void hnormalize(const double *x, const int rows, const int npts, double *nx);

static double mat_inverse3x3(const double h[9], double invh[9]);
static bool iscolinear(const double *p0, const double *p1, const double *p2);
static char signofdet3(const double *p0, const double *p1, const double *p2);
static double determinant3(const double *p0, const double *p1, const double *p2);

/*
template <ProjectiveTransformation::WarpModel w, DegeneracyCheck::Level c > 
bool isdegenerate(const double *x, const int m);

template <ProjectiveTransformation::WarpModel w> 
int homography2d(double *pts, const int rows, const int npts, double H[9]);
*/

bool isdegenerate(ProjectiveTransformation::WarpModel w, DegeneracyCheck::Level c, const double *x, const int m);
bool isdegenerate_homography_0(const double *x, const int m);
bool isdegenerate_affine_0(const double *x, const int m);
bool isdegenerate_similarity_0(const double *x, const int m);
bool isdegenerate_homography_1(const double *x, const int m);
bool isdegenerate_affine_1(const double *x, const int m);
bool isdegenerate_homography_2(const double *x, const int m);
bool isdegenerate_affine_2(const double *x, const int m);
bool isdegenerate_homography_3(const double *x, const int m);
bool isdegenerate_affine_3(const double *x, const int m);


int homography2d(ProjectiveTransformation::WarpModel w, double *pts, const int rows, const int npts, double H[9]);
int homography2d_homography(double *pts, const int rows, const int npts, double H[9]);
int homography2d_affine(double *pts, const int rows, const int npts, double H[9]);
int homography2d_similarity(double *pts, const int rows, const int npts, double H[9]);

int homogdist2d(const double H[9], const double *x, const int rows, const int npts, const double t, int *inliers);

template <ParallelCheck::Level c>
bool isnonparallel(const double *nx, const int rows, const int *inliers, const int ninliers,
                   const double costhresh = ParallelCheck::cosanglethresh, 
                   const double voffset = ParallelCheck::vertoffset,
                   const int ntrials = ParallelCheck::ntrials);

bool invalidhomography(const double H[9]);

#endif
