
#ifndef _OBJECTTRACKER_H
#define _OBJECTTRACKER_H

#include "TheTracker.h"
#include <list>
#include "ObjectModel.h"
#include "RegionSegmentation.h"
#include "OcclusionHandler.h"
#include "ContourExtractor.h"

class ObjectTracker
{
public:
	enum ContourMethod {
		LEVEL_SETS = 0,
		FAST_LEVEL_SETS = 1,
		WALL_FOLLOW = 2
	};
private:

	// The core function that computes the strength map
	void _ComputeStrengthMap(const ImgBgr& current_frame);
	void _ComputeSpatialStrength();

	// Computes the motion using Joint-KLT and updates the spatial model of the object.
	void _ComputeMotion(const ImgBgr& current_frame);

	// Display the reconstrucuted object. Should be used only for display purpose.
	void _DisplayReconstructedScene();
	
	// Handlers to all objects
	OcclusionHandler m_occlusion_handler;
	ObjectModel m_object_model;
	ContourExtractor m_contour_extractor;
	RegionSegmentation m_segmentation_handler;

	// General variables 
	int m_width, m_height;
	int m_frame_no, m_frame_count;
	int m_distance_threshold;
	int m_num_features;
	Point m_object_centroid;
	double m_seg_sigma;

	// Strength images..
	ImgFloat m_fg_strength;
	ImgFloat m_bg_strength;
	ImgFloat m_strength_image;   // Stores the log ratio of the fg and bg prob
	ImgBinary m_object_mask;

	// Motion parameters
	ImgBgr m_prev_frame;
	int m_nlevels;
	int m_mindist;

	// Contour info
	Bgr m_contour_color;
	int m_obj_id;

	// Parameters that control the algorithm
	bool m_use_motion;
	bool m_use_region_consistency;
	bool m_use_edges;

	ContourMethod m_contour_method;

	void _ComputeDecisionBoundary(const ImgBgr& current_frame);

public:

	ObjectTracker(int width, int height);
	~ObjectTracker();
	int Initialize(const ImgBgr &input_image, vector<Point>& points, int obj_id, double seg_sigma = 3.0f, int num_features = 3, ImgInt *const fg_label = NULL, ImgInt *const bg_label = NULL);
	int TrackNextFrame(ImgBgr &current_frame, int frame_no);
	void DrawContour(ImgBgr &display_image);
	void LoadStateFromFile(const ImgInt& current_model);
};

#endif