#include <iostream>
#include <stdlib.h>
#include <cmath>
#include "blepo.h"
#include "vector"
#include "ScanfillSkeleton.h"


void insertEdge(EdgePolygon *list, EdgePolygon *edge) 
{
	EdgePolygon *p, *q = list;
	p = q->next;
	while(p != NULL) {
		if(edge->xIntersect < p->xIntersect)
			p = NULL;
		else {
			q = p;
			p = p->next;
		}
	}
	edge->next = q ->next;
	q->next = edge;
}

int gblindex = 0;
void makeEdgeRec(iPoint lower, iPoint upper, int yComp, EdgePolygon *edge, EdgePolygon *edges[]) 
{
	edge->dxPerScan = (float) (upper.col - lower.col) / (upper.row - lower.row);
	edge->xIntersect = lower.col;
	if (upper.row < yComp){
		edge->yUpper = upper.row - 1;
		edge->flag = true;
	}
	else{
		edge->yUpper = upper.row;
		edge->flag = false;
	}
	insertEdge(edges[lower.row], edge);
}


void buildEdgeList(int cnt, iPoint *pts, EdgePolygon *edges[]) 
{
	EdgePolygon *edge;
	iPoint v1, v2;
	int i, yPrev = pts[cnt-2].row;

	/* Initialize the process with the last point (connects to the first point) */
	v1.col = pts[cnt-1].col;
	v1.row = pts[cnt-1].row;

	/* for each line segment */
	for(i=0;i<cnt;i++) {

		/* end of the segment is the curernt point */
		v2.col = pts[i].col;
		v2.row = pts[i].row;
		if(v1.row != v2.row) 
		{ /* not a horizontal line */

			/* create a new edge with v1.row less than v2.row */
			edge = (EdgePolygon *)malloc(sizeof(EdgePolygon));
			if(v1.row < v2.row) 
				makeEdgeRec(v1, v2, 0, edge, edges);
			else
				makeEdgeRec(v2, v1, yPrev, edge, edges);
		}
		yPrev = v1.row;
		v1 = v2;
	}

}


void buildActiveList(int scan, EdgePolygon *active, EdgePolygon *edges[]) 
{
	EdgePolygon *p, *q;

	p = edges[scan]->next;
	while(p) 
	{
		q = p->next;
		insertEdge(active, p);
		p = q;
	}
}

void fillScan(int scan, EdgePolygon *active,ImgBinary *im,int rows,int cols) 
{
	EdgePolygon *p1, *p2;
	int i, f;

	p1 = active->next;
	//p1 = active;
	while(p1!=NULL&&p1->next!=NULL) 
	{
		p2 = p1->next;

		for (i=p1->xIntersect; i<p2->xIntersect; i++)
			(*im)(i,scan) = 1;
		p1 = p2->next;
	}

}

void deleteAfter(EdgePolygon *q) 
{
	EdgePolygon *p = q->next;

	q->next = p->next;
	free(p);
}

void updateActiveList(int scan, EdgePolygon *active) 
{
	EdgePolygon *q = active, *p = active->next;
	int val;
	while(p)
	{
		/* if the edge has ended, get rid of it */
		if(p->flag)
			val=p->yUpper;
		else
			val=p->yUpper-1;
		if(scan >= val) 
		{
			p = p->next;
			deleteAfter(q);
		}
		/*  otherwise, update the xIntersect value */
		else 
		{
			p->xIntersect = p->xIntersect + p->dxPerScan;
			q = p;
			p = p->next;
		}
	}
}


void resortActiveList(EdgePolygon *active) 
{
	EdgePolygon *q, *p = active->next;

	active->next = NULL;
	while(p) 
	{
		q = p->next;
		insertEdge(active, p);
		p = q;
	}
}

int FillPolygon( ImgBinary *img,std::vector<Point> points)
{
	//unsigned char * img = calloc(sizeof(unsigned char),rows*cols);
	EdgePolygon **edges, *edgeData, *active;
	int i, scan,cnt;
	cnt = points.size();
	iPoint * p = (iPoint *)malloc(sizeof(iPoint)*cnt);
	//int width = cols;
	//int height = rows;
	int rows = img->Height();
	int cols = img->Width();

	for(i=0;i<cnt;i++)
	{
		p[i].col = points[i].x;
		p[i].row = points[i].y;
	}


	edges = (EdgePolygon **)malloc(sizeof(EdgePolygon *) * rows); //rows
	if(edges == NULL)
	{
		printf("Allocation error\n");
		exit(0);
	}

	for(i=0;i<rows;i++) //rows
	{
		edges[i] = (EdgePolygon *)malloc(sizeof(EdgePolygon));
		edges[i]->next = NULL;
	}

	buildEdgeList(cnt, p, edges);
	active = (EdgePolygon *)malloc(sizeof(EdgePolygon));
	active->next = NULL;

	for(scan = 0; scan < rows; scan++) //rows
	{
		buildActiveList(scan, active, edges);
		if(active->next) 
		{
			fillScan(scan, active, img ,rows,cols);
			updateActiveList(scan, active);
			resortActiveList(active);
		}
	}

	/* free malloced edges! */
	for(i=0;i<rows;i++)
		free(edges[i]);

	free(edges);
	return 1;
}
