
#include "Utils.h"

#define BIG 100000



void Utils::Inverse2x2(double input_mat[2][2], double out[2][2])
{
	double det = (input_mat[1][1]*input_mat[0][0]) - (input_mat[0][1]*input_mat[1][0]);
	if(det == 0) {
		det = 0.01f;
		printf("Warning: 2x2 determinant 0\n");
	}
	out[0][0] = input_mat[1][1]/det;
	out[0][1] = -input_mat[0][1]/det;
	out[1][0] = -input_mat[1][0]/det;
	out[1][1] = input_mat[0][0]/det;
}

bool Utils::Inverse3x3(double input_mat[3][3], double out[3][3])
{
	bool ret = true;
	double det = (input_mat[0][0] * ((input_mat[2][2]*input_mat[1][1]) - (input_mat[2][1]*input_mat[1][2]))) - (input_mat[1][0] * ((input_mat[2][2]*input_mat[0][1]) - (input_mat[2][1]*input_mat[0][2]))) + (input_mat[2][0] * ((input_mat[1][2]*input_mat[0][1]) - (input_mat[1][1]*input_mat[0][2])));
	if(det == 0) {
		det = 0.01f;
		printf("Warning: 3x3 determinant is 0\n");
		ret = false;
	}
	out[0][0] = ((input_mat[2][2]*input_mat[1][1]) - (input_mat[2][1]*input_mat[1][2]))/det;
	out[0][1] = -((input_mat[2][2]*input_mat[0][1]) - (input_mat[2][1]*input_mat[0][2]))/det;
	out[0][2] = ((input_mat[1][2]*input_mat[0][1]) - (input_mat[1][1]*input_mat[0][2]))/det;
	out[1][0] = -((input_mat[2][2]*input_mat[1][0]) - (input_mat[2][0]*input_mat[1][2]))/det;
	out[1][1] = ((input_mat[2][2]*input_mat[0][0]) - (input_mat[2][0]*input_mat[0][2]))/det;
	out[1][2] = -((input_mat[1][2]*input_mat[0][0]) - (input_mat[1][0]*input_mat[0][2]))/det;
	out[2][0] = ((input_mat[2][1]*input_mat[1][0]) - (input_mat[2][0]*input_mat[1][1]))/det;
	out[2][1] = -((input_mat[2][1]*input_mat[0][0]) - (input_mat[2][0]*input_mat[0][1]))/det;
	out[2][2] = ((input_mat[1][1]*input_mat[0][0]) - (input_mat[1][0]*input_mat[0][1]))/det;
	return ret;
}

void Utils::RandomColor(const ImgInt& img, ImgBgr* out)
{
  // convert to pseudo-random colors
  out->Reset(img.Width(), img.Height());
  ImgInt::ConstIterator p;
  ImgBgr::Iterator q;
  for (p = img.Begin(), q = out->Begin() ; p != img.End() ; p++, q++)
  {
    // There is no significance to the integer here
	  if(*p > 0)
    *q = ImgBgr::Pixel((*p * 12342223) % 0xFFFFFF, Bgr::BLEPO_BGR_XBGR);
	  //else if(*p < 0)
		//*q = Bgr(255,255,255);
	  else
			*q = Bgr(0,0,0);
  }
}

void Utils::RandomColor(const ImgGray& img, ImgBgr* out)
{
  // convert to pseudo-random colors
  out->Reset(img.Width(), img.Height());
  ImgGray::ConstIterator p;
  ImgBgr::Iterator q;
  for (p = img.Begin(), q = out->Begin() ; p != img.End() ; p++, q++)
  {
    // There is no significance to the integer here
	  if(*p > 0)
	    *q = ImgBgr::Pixel((*p * 12342223) % 0xFFFFFF, Bgr::BLEPO_BGR_XBGR);
	  //else if(*p < 0)
		//*q = Bgr(255,255,255);
	  else
			*q = Bgr(0,0,0);
  }
}

void Utils:: _EdgeIndicator(const ImgFloat &Ix,const ImgFloat &Iy,ImgFloat *g)
{
	g->Reset(Ix.Width(),Ix.Height());
	ImgFloat::ConstIterator I1 = Ix.Begin();
	ImgFloat::ConstIterator I2 = Iy.Begin();
	ImgFloat::Iterator out = g->Begin();	
	for ( ; out != g->End() ; out++,I1++,I2++)  
		*out = 1/(1+((*I1)*(*I1) + (*I2)*(*I2)));
}

void Utils::_ExtractContour(const ImgFloat &in,ImgInt *labels,int *area,vector<Point> *chain,boolean IsWallfollow)
{
	int count = 0;
	
	labels->Reset(in.Width(),in.Height());
	ImgFloat::ConstIterator p1 = in.Begin();
	ImgInt::Iterator lp = labels->Begin();
	for ( ; p1 != in.End();){
		if(*p1++ <=0 ){
			count++;
			*lp++ = 1;
		}
		else
			*lp++ = 0;
	}

	*area = count;

	if(IsWallfollow)
		WallFollow(*labels,1,chain);
}

void Utils::HausdorffDistance(vector<Point> &p1, vector<Point> &p2,double *dist)
{
	int m = p1.size(),k,l;
	int n = p2.size();
	int dim = 2;
	int Cx,Cy,Dx,Dy;
	double D,minD=0,maxDist=9999;
	for(k=0;k<m;k++)
	{
		Cx = p1[k].x;Cy = p1[k].y;minD=99999;
		for(l=0;l<n;l++)
		{
			Dx = (Cx-p2[l].x)*(Cx-p2[l].x);
			Dy = (Cy-p2[l].y)*(Cy-p2[l].y);
			D = sqrt(Dx+Dy);
			if(l==0 || D<minD)
				minD = D;
		}
		
		if(k==0 || minD>maxDist)
			maxDist = minD;
	}
	*dist = maxDist;
}

void Utils::ConvertStrengthtoBinary(const ImgFloat& img, ImgBinary* out)
{
  out->Reset(img.Width(), img.Height());
  ImgFloat::ConstIterator p;
  ImgBinary::Iterator q;
  for (p = img.Begin(), q = out->Begin() ; p != img.End() ; p++, q++)
  {
    *q = (*p>0);
  }
}

void Utils::SetLevelSetSpeed(const ImgFloat& img, ImgFloat* out)
{
 ImgFloat::ConstIterator p = img.Begin();
 ImgFloat::Iterator q = out->Begin();
 for (;p != img.End() ; p++, q++)
 {
  if(*p <= 0){
   //m_strength_image(x,y) = 0;
   *q = 1.5;
  }
  else
   *q = -1.5;
 }
}

double Utils::lap(int dim,double **assigncost,int *rowsol,int *colsol,double *u,double *v)
{
  int unassignedfound;
  double min,umin,usubmin,h,v2,*d;
  int imin,j,j2,i,numfree=0,prvnumfree,*free,*pred,k,f,j1,*matches,i0,freerow,*collist,up,low,last,endofpath;

  free = new int[dim];       // list of unassigned rows.
  collist = new int[dim];    // list of columns to be scanned in various ways.
  matches = new int[dim];    // counts how many times a row could be assigned.
  d = new double[dim];         // 'cost-distance' in augmenting path calculation.
  pred = new int[dim];       // row-predecessor of column in augmenting/alternating path.

  // init how many times a row will be assigned in the column reduction.
  for (i = 0; i < dim; i++)  
    matches[i] = 0;

  // COLUMN REDUCTION 
  for (j = dim-1; j >= 0; j--)    // reverse order gives better results.
  {
    // find minimum cost over rows.
    min = assigncost[0][j]; 
    imin = 0;
    for (i = 1; i < dim; i++)  
      if (assigncost[i][j] < min) 
      { 
        min = assigncost[i][j]; 
        imin = i;
      }
    v[j] = min; 

    if (++matches[imin] == 1) 
    { 
      // init assignment if minimum row assigned for first time.
      rowsol[imin] = j; 
      colsol[j] = imin; 
    }
    else
      colsol[j] = -1;        // row already assigned, column not assigned.
  }

  // REDUCTION TRANSFER
  for (i = 0; i < dim; i++) 
    if (matches[i] == 0)     // fill list of unassigned 'free' rows.
      free[numfree++] = i;
    else
      if (matches[i] == 1)   // transfer reduction from rows that are assigned once.
      {
        j1 = rowsol[i]; 
        min = BIG;
        for (j = 0; j < dim; j++)  
          if (j != j1)
            if (assigncost[i][j] - v[j] < min) 
              min = assigncost[i][j] - v[j];
        v[j1] = v[j1] - min;
      }

  // AUGMENTING ROW REDUCTION 
  int loopcnt = 0;           // do-loop to be done twice.
  do
  {
    loopcnt++;

    // scan all free rows.
    // in some cases, a free row may be replaced with another one to be scanned next.
    k = 0; 
    prvnumfree = numfree; 
    numfree = 0;             // start list of rows still free after augmenting row reduction.
    while (k < prvnumfree)
    {
      i = free[k]; 
      k++;

      // find minimum and second minimum reduced cost over columns.
      umin = assigncost[i][0] - v[0]; 
      j1 = 0; 
      usubmin = BIG;
      for (j = 1; j < dim; j++) 
      {
        h = assigncost[i][j] - v[j];
        if (h < usubmin)
          if (h >= umin) 
          { 
            usubmin = h; 
            j2 = j;
          }
          else 
          { 
            usubmin = umin; 
            umin = h; 
            j2 = j1; 
            j1 = j;
          }
      }

      i0 = colsol[j1];
      if (umin < usubmin) 
        // change the reduction of the minimum column to increase the minimum
        // reduced cost in the row to the subminimum.
        v[j1] = v[j1] - (usubmin - umin);
      else                   // minimum and subminimum equal.
        if (i0 >= 0)         // minimum column j1 is assigned.
        { 
          // swap columns j1 and j2, as j2 may be unassigned.
          j1 = j2; 
          i0 = colsol[j2];
        }

      // (re-)assign i to j1, possibly de-assigning an i0.
      rowsol[i] = j1; 
      colsol[j1] = i;

      if (i0 >= 0)           // minimum column j1 assigned earlier.
        if (umin < usubmin) 
          // put in current k, and go back to that k.
          // continue augmenting path i - j1 with i0.
          free[--k] = i0; 
        else 
          // no further augmenting reduction possible.
          // store i0 in list of free rows for next phase.
          free[numfree++] = i0; 
    }
  }
  while (loopcnt < 2);       // repeat once.

  // AUGMENT SOLUTION for each free row.
  for (f = 0; f < numfree; f++) 
  {
    freerow = free[f];       // start row of augmenting path.

    // Dijkstra shortest path algorithm.
    // runs until unassigned column added to shortest path tree.
    for (j = 0; j < dim; j++)  
    { 
      d[j] = assigncost[freerow][j] - v[j]; 
      pred[j] = freerow;
      collist[j] = j;        // init column list.
    }

    low = 0; // columns in 0..low-1 are ready, now none.
    up = 0;  // columns in low..up-1 are to be scanned for current minimum, now none.
             // columns in up..dim-1 are to be considered later to find new minimum, 
             // at this stage the list simply contains all columns 
    unassignedfound = FALSE;
    do
    {
      if (up == low)         // no more columns to be scanned for current minimum.
      {
        last = low - 1; 

        // scan columns for up..dim-1 to find all indices for which new minimum occurs.
        // store these indices between low..up-1 (increasing up). 
        min = d[collist[up++]]; 
        for (k = up; k < dim; k++) 
        {
          j = collist[k]; 
          h = d[j];
          if (h <= min)
          {
            if (h < min)     // new minimum.
            { 
              up = low;      // restart list at index low.
              min = h;
            }
            // new index with same minimum, put on undex up, and extend list.
            collist[k] = collist[up]; 
            collist[up++] = j; 
          }
        }

        // check if any of the minimum columns happens to be unassigned.
        // if so, we have an augmenting path right away.
        for (k = low; k < up; k++) 
          if (colsol[collist[k]] < 0) 
          {
            endofpath = collist[k];
            unassignedfound = TRUE;
            break;
          }
      }

      if (!unassignedfound) 
      {
        // update 'distances' between freerow and all unscanned columns, via next scanned column.
        j1 = collist[low]; 
        low++; 
        i = colsol[j1]; 
        h = assigncost[i][j1] - v[j1] - min;

        for (k = up; k < dim; k++) 
        {
          j = collist[k]; 
          v2 = assigncost[i][j] - v[j] - h;
          if (v2 < d[j])
          {
            pred[j] = i;
            if (v2 == min)   // new column found at same minimum value
              if (colsol[j] < 0) 
              {
                // if unassigned, shortest augmenting path is complete.
                endofpath = j;
                unassignedfound = TRUE;
                break;
              }
              // else add to list to be scanned right away.
              else 
              { 
                collist[k] = collist[up]; 
                collist[up++] = j; 
              }
            d[j] = v2;
          }
        }
      } 
    }
    while (!unassignedfound);

    // update column prices.
    for (k = 0; k <= last; k++)  
    { 
      j1 = collist[k]; 
      v[j1] = v[j1] + d[j1] - min;
    }

    // reset row and column assignments along the alternating path.
    do
    {
      i = pred[endofpath]; 
      colsol[endofpath] = i; 
      j1 = endofpath; 
      endofpath = rowsol[i]; 
      rowsol[i] = j1;
    }
    while (i != freerow);
  }

  // calculate optimal cost.
  double lapcost = 0;
  for (i = 0; i < dim; i++)  
  {
    j = rowsol[i];
    u[i] = assigncost[i][j] - v[j];
    lapcost = lapcost + assigncost[i][j]; 
  }

  // free reserved memory.
  delete[] pred;
  delete[] free;
  delete[] collist;
  delete[] matches;
  delete[] d;

  return lapcost;
}